@@ -1,9 +1,40 @@
-#This file is part of Tryton.  The COPYRIGHT file at the top level of
-#this repository contains the full copyright notices and license terms.
+# This file is part of Tryton.  The COPYRIGHT file at the top level of
+# this repository contains the full copyright notices and license terms.
 
+import re
 import operator
-import types
 import datetime
+from collections import defaultdict
+from functools import reduce, partial
+
+
+def sql_like(value, pattern, ignore_case=True):
+    flag = re.IGNORECASE if ignore_case else 0
+
+    escape = False
+    chars = []
+    for char in re.split(r'(\\|.)', pattern)[1::2]:
+        if escape:
+            if char in ('%', '_'):
+                chars.append(char)
+            else:
+                chars.extend(['\\', char])
+            escape = False
+        elif char == '\\':
+            escape = True
+        elif char == '_':
+            chars.append('.')
+        elif char == '%':
+            chars.append('.*')
+        else:
+            chars.append(re.escape(char))
+
+    regexp = re.compile(''.join(chars), flag)
+    return bool(regexp.fullmatch(value))
+
+
+like = partial(sql_like, ignore_case=False)
+ilike = partial(sql_like, ignore_case=True)
 
 
 def in_(a, b):
@@ -15,23 +46,22 @@ def in_(a, b):
     else:
         return operator.contains(b, a)
 
-OPERATORS = {
-    '=': operator.eq,
-    '>': operator.gt,
-    '<': operator.lt,
-    '<=': operator.le,
-    '>=': operator.ge,
-    '!=': operator.ne,
-    'in': in_,
-    'not in': lambda a, b: not in_(a, b),
-    # Those operators are not supported (yet ?)
-    'like': lambda a, b: True,
-    'ilike': lambda a, b: True,
-    'not like': lambda a, b: True,
-    'not ilike': lambda a, b: True,
-    'child_of': lambda a, b: True,
-    'not child_of': lambda a, b: True,
-}
+
+OPERATORS = defaultdict(lambda: lambda a, b: True)
+OPERATORS.update({
+        '=': operator.eq,
+        '>': operator.gt,
+        '<': operator.lt,
+        '<=': operator.le,
+        '>=': operator.ge,
+        '!=': operator.ne,
+        'in': in_,
+        'not in': lambda a, b: not in_(a, b),
+        'ilike': ilike,
+        'not ilike': lambda a, b: not ilike(a, b),
+        'like': like,
+        'not like': lambda a, b: not like(a, b),
+        })
 
 
 def locale_part(expression, field_name, locale_name='id'):
@@ -46,8 +76,7 @@ def locale_part(expression, field_name, locale_name='id'):
 def is_leaf(expression):
     return (isinstance(expression, (list, tuple))
         and len(expression) > 2
-        and isinstance(expression[1], basestring)
-        and expression[1] in OPERATORS)
+        and isinstance(expression[1], str))
 
 
 def constrained_leaf(part, boolop=operator.and_):
@@ -76,21 +105,27 @@ def eval_leaf(part, context, boolop=operator.and_):
             context_field = datetime.datetime.min
         else:
             context_field = datetime.date.min
-    if (isinstance(context_field, basestring)
+    if isinstance(context_field, (list, tuple)) and value is None:
+        value = type(context_field)()
+    if (isinstance(context_field, str)
             and isinstance(value, (list, tuple))):
         try:
-            value = '%s,%s' % value
+            value = '%s,%s' % tuple(value)
         except TypeError:
             pass
     elif (isinstance(context_field, (list, tuple))
-            and isinstance(value, basestring)):
+            and isinstance(value, str)):
         try:
-            context_field = '%s,%s' % context_field
+            context_field = '%s,%s' % tuple(context_field)
         except TypeError:
             pass
+    elif ((isinstance(context_field, list) and isinstance(value, tuple))
+            or (isinstance(context_field, tuple) and isinstance(value, list))):
+        context_field = list(context_field)
+        value = list(value)
     if (operand in ('=', '!=')
             and isinstance(context_field, (list, tuple))
-            and isinstance(value, (int, long))):
+            and isinstance(value, int)):
         operand = {
             '=': 'in',
             '!=': 'not in',
@@ -112,7 +147,49 @@ def inverse_leaf(domain):
                 return [domain[3]] + list(domain[1:])
         return domain
     else:
-        return map(inverse_leaf, domain)
+        return list(map(inverse_leaf, domain))
+
+
+def filter_leaf(domain, field, model):
+    if domain in ('AND', 'OR'):
+        return domain
+    elif is_leaf(domain):
+        if domain[0].startswith(field) and len(domain) > 3:
+            if domain[3] != model:
+                return ('id', '=', None)
+        return domain
+    else:
+        return [filter_leaf(d, field, model) for d in domain]
+
+
+def prepare_reference_domain(domain, reference):
+    "convert domain to replace reference fields by their local part"
+    if domain in ('AND', 'OR'):
+        return domain
+    elif is_leaf(domain):
+        # When a Reference field is using the dotted notation the model
+        # specified must be removed from the clause
+        if domain[0].count('.') and len(domain) > 3:
+            local_name, target_name = domain[0].split('.', 1)
+            if local_name == reference:
+                return [target_name] + list(domain[1:3] + domain[4:])
+        return domain
+    else:
+        return [prepare_reference_domain(d, reference) for d in domain]
+
+
+def extract_reference_models(domain, field_name):
+    "returns the set of the models available for field_name"
+    if domain in ('AND', 'OR'):
+        return set()
+    elif is_leaf(domain):
+        local_part = domain[0].split('.', 1)[0]
+        if local_part == field_name and len(domain) > 3:
+            return {domain[3]}
+        return set()
+    else:
+        return reduce(operator.or_,
+            (extract_reference_models(d, field_name) for d in domain))
 
 
 def eval_domain(domain, context, boolop=operator.and_):
@@ -132,23 +209,28 @@ def eval_domain(domain, context, boolop=operator.and_):
             eval_domain(domain[1:], context, boolop))
 
 
-def localize_domain(domain, field_name=None):
+def localize_domain(domain, field_name=None, strip_target=False):
     "returns only locale part of domain. eg: langage.code -> code"
     if domain in ('AND', 'OR', True, False):
         return domain
     elif is_leaf(domain):
         if 'child_of' in domain[1]:
+            if domain[0].count('.'):
+                _, target_part = domain[0].split('.', 1)
+                return [target_part] + list(domain[1:])
             if len(domain) == 3:
                 return domain
             else:
                 return [domain[3]] + list(domain[1:-1])
         locale_name = 'id'
-        if isinstance(domain[2], basestring):
+        if isinstance(domain[2], str):
             locale_name = 'rec_name'
+        n = 3 if strip_target else 4
         return [locale_part(domain[0], field_name, locale_name)] \
-            + list(domain[1:3]) + list(domain[4:])
+            + list(domain[1:n]) + list(domain[4:])
     else:
-        return [localize_domain(part, field_name) for part in domain]
+        return [localize_domain(part, field_name, strip_target)
+            for part in domain]
 
 
 def simplify(domain):
@@ -196,6 +278,17 @@ def concat(*domains, **kwargs):
     return simplify(merge(result))
 
 
+def unique_value(domain):
+    "Return if unique, the field and the value"
+    if (isinstance(domain, list)
+            and len(domain) == 1
+            and '.' not in domain[0][0]
+            and domain[0][1] == '='):
+        return True, domain[0][1], domain[0][2]
+    else:
+        return False, None, None
+
+
 def parse(domain):
     if is_leaf(domain):
         return domain
@@ -221,7 +314,7 @@ def domain_inversion(domain, symbol, context=None):
 class And(object):
 
     def __init__(self, expressions):
-        self.branches = map(parse, expressions)
+        self.branches = list(map(parse, expressions))
         self.variables = set()
         for expression in self.branches:
             if is_leaf(expression):
@@ -240,7 +333,7 @@ class And(object):
         for part in self.branches:
             if isinstance(part, And):
                 part_inversion = part.inverse(symbol, context)
-                evaluated = isinstance(part_inversion, types.BooleanType)
+                evaluated = isinstance(part_inversion, bool)
                 if not evaluated:
                     result.append(part_inversion)
                 elif part_inversion:
@@ -259,7 +352,7 @@ class And(object):
                 else:
                     return False
 
-        result = filter(lambda e: e is not True, result)
+        result = [e for e in result if e is not True]
         if result == []:
             return True
         else:
@@ -280,7 +373,7 @@ class Or(And):
         for part in self.branches:
             if isinstance(part, And):
                 part_inversion = part.inverse(symbol, context)
-                evaluated = isinstance(part_inversion, types.BooleanType)
+                evaluated = isinstance(part_inversion, bool)
                 if symbol not in part.variables:
                     if evaluated and part_inversion:
                         return True
@@ -298,13 +391,13 @@ class Or(And):
                 field = self.base(field)
                 if (field in context
                         and (eval_leaf(part, context, operator.or_)
-                                or constrained_leaf(part, operator.or_))):
+                            or constrained_leaf(part, operator.or_))):
                     return True
                 elif (field in context
                         and not eval_leaf(part, context, operator.or_)):
                     result.append(False)
 
-        result = filter(lambda e: e is not False, result)
+        result = [e for e in result if e is not False]
         if result == []:
             return False
         else:
@@ -365,7 +458,7 @@ def test_or_inversion():
     assert domain_inversion(domain, 'y', {'z': 4}) == [['y', '<', 3]]
     assert domain_inversion(domain, 'y', {'x': 3}) is True
 
-    domain = [u'OR', [u'length', u'>', 5], [u'language.code', u'=', u'de_DE']]
+    domain = ['OR', ['length', '>', 5], ['language.code', '=', 'de_DE']]
     assert domain_inversion(domain, 'length', {'length': 0, 'name': 'n'}) ==\
         [['length', '>', 5]]
 
@@ -484,6 +577,17 @@ def test_concat():
         'OR', [['a', '=', 1]], [['b', '=', 2]]]
 
 
+def test_unique_value():
+    domain = [['a', '=', 1]]
+    assert unique_value(domain) == (True, '=', 1)
+    domain = [['a', '!=', 1]]
+    assert unique_value(domain)[0] is False
+    domain = [['a', '=', 1], ['a', '=', 2]]
+    assert unique_value(domain)[0] is False
+    domain = [['a.b', '=', 1]]
+    assert unique_value(domain)[0] is False
+
+
 def test_evaldomain():
     domain = [['x', '>', 5]]
     assert eval_domain(domain, {'x': 6})
@@ -512,6 +616,59 @@ def test_evaldomain():
     assert not eval_domain(domain, {'x': [3, 4]})
     assert eval_domain(domain, {'x': [1, 2]})
 
+    domain = [['x', 'like', 'abc']]
+    assert eval_domain(domain, {'x': 'abc'})
+    assert not eval_domain(domain, {'x': ''})
+    assert not eval_domain(domain, {'x': 'xyz'})
+    assert not eval_domain(domain, {'x': 'abcd'})
+
+    domain = [['x', 'not like', 'abc']]
+    assert eval_domain(domain, {'x': 'xyz'})
+    assert eval_domain(domain, {'x': 'ABC'})
+    assert not eval_domain(domain, {'x': 'abc'})
+
+    domain = [['x', 'not ilike', 'abc']]
+    assert eval_domain(domain, {'x': 'xyz'})
+    assert not eval_domain(domain, {'x': 'ABC'})
+    assert not eval_domain(domain, {'x': 'abc'})
+
+    domain = [['x', 'like', 'a%']]
+    assert eval_domain(domain, {'x': 'a'})
+    assert eval_domain(domain, {'x': 'abcde'})
+    assert not eval_domain(domain, {'x': ''})
+    assert not eval_domain(domain, {'x': 'ABCDE'})
+    assert not eval_domain(domain, {'x': 'xyz'})
+
+    domain = [['x', 'ilike', 'a%']]
+    assert eval_domain(domain, {'x': 'a'})
+    assert eval_domain(domain, {'x': 'A'})
+    assert not eval_domain(domain, {'x': ''})
+    assert not eval_domain(domain, {'x': 'xyz'})
+
+    domain = [['x', 'like', 'a_']]
+    assert eval_domain(domain, {'x': 'ab'})
+    assert not eval_domain(domain, {'x': 'a'})
+    assert not eval_domain(domain, {'x': 'abc'})
+
+    domain = [['x', 'like', 'a\\%b']]
+    assert eval_domain(domain, {'x': 'a%b'})
+    assert not eval_domain(domain, {'x': 'ab'})
+    assert not eval_domain(domain, {'x': 'a123b'})
+
+    domain = [['x', 'like', '\\%b']]
+    assert eval_domain(domain, {'x': '%b'})
+    assert not eval_domain(domain, {'x': 'b'})
+    assert not eval_domain(domain, {'x': '123b'})
+
+    domain = [['x', 'like', 'a\\_c']]
+    assert eval_domain(domain, {'x': 'a_c'})
+    assert not eval_domain(domain, {'x': 'abc'})
+    assert not eval_domain(domain, {'x': 'ac'})
+
+    domain = [['x', 'like', 'a\\\\_c']]
+    assert eval_domain(domain, {'x': 'a\\bc'})
+    assert not eval_domain(domain, {'x': 'abc'})
+
     domain = ['OR', ['x', '>', 10], ['x', '<', 0]]
     assert eval_domain(domain, {'x': 11})
     assert eval_domain(domain, {'x': -4})
@@ -554,6 +711,22 @@ def test_evaldomain():
     assert eval_domain(domain, {'x': [1, 2]})
     assert not eval_domain(domain, {'x': [2]})
 
+    domain = [['x', '=', None]]
+    assert eval_domain(domain, {'x': []})
+
+    domain = [['x', '=', ['foo', 1]]]
+    assert eval_domain(domain, {'x': 'foo,1'})
+    assert eval_domain(domain, {'x': ('foo', 1)})
+    assert eval_domain(domain, {'x': ['foo', 1]})
+    domain = [['x', '=', ('foo', 1)]]
+    assert eval_domain(domain, {'x': 'foo,1'})
+    assert eval_domain(domain, {'x': ('foo', 1)})
+    assert eval_domain(domain, {'x': ['foo', 1]})
+
+    domain = [['x', '=', 'foo,1']]
+    assert eval_domain(domain, {'x': ['foo', 1]})
+    assert eval_domain(domain, {'x': ('foo', 1)})
+
 
 def test_localize():
     domain = [['x', '=', 5]]
@@ -576,11 +749,49 @@ def test_localize():
     domain = [['x', 'child_of', [1], 'y']]
     assert localize_domain(domain, 'x') == [['y', 'child_of', [1]]]
 
+    domain = [['x.y', 'child_of', [1], 'parent']]
+    assert localize_domain(domain, 'x') == [['y', 'child_of', [1], 'parent']]
+
+    domain = [['x.y.z', 'child_of', [1], 'parent', 'model']]
+    assert localize_domain(domain, 'x') == \
+        [['y.z', 'child_of', [1], 'parent', 'model']]
+
     domain = [['x.id', '=', 1, 'y']]
-    assert localize_domain(domain, 'x') == [['id', '=', 1]]
+    assert localize_domain(domain, 'x', False) == [['id', '=', 1, 'y']]
+    assert localize_domain(domain, 'x', True) == [['id', '=', 1]]
 
     domain = [['a.b.c', '=', 1, 'y', 'z']]
-    assert localize_domain(domain, 'x') == [['b.c', '=', 1, 'z']]
+    assert localize_domain(domain, 'x', False) == [['b.c', '=', 1, 'y', 'z']]
+    assert localize_domain(domain, 'x', True) == [['b.c', '=', 1, 'z']]
+
+
+def test_prepare_reference_domain():
+    domain = [['x', 'like', 'A%']]
+    assert prepare_reference_domain(domain, 'x') == [['x', 'like', 'A%']]
+
+    domain = [['x.y', 'like', 'A%', 'model']]
+    assert prepare_reference_domain(domain, 'x') == [['y', 'like', 'A%']]
+
+    domain = [['x.y', 'child_of', [1], 'model', 'parent']]
+    assert prepare_reference_domain(domain, 'x') == \
+        [['y', 'child_of', [1], 'parent']]
+
+
+def test_extract_models():
+    domain = [['x', 'like', 'A%']]
+    assert extract_reference_models(domain, 'x') == set()
+    assert extract_reference_models(domain, 'y') == set()
+
+    domain = [['x', 'like', 'A%', 'model']]
+    assert extract_reference_models(domain, 'x') == {'model'}
+    assert extract_reference_models(domain, 'y') == set()
+
+    domain = ['OR',
+        ['x.y', 'like', 'A%', 'model_A'],
+        ['x.z', 'like', 'B%', 'model_B']]
+    assert extract_reference_models(domain, 'x') == {'model_A', 'model_B'}
+    assert extract_reference_models(domain, 'y') == set()
+
 
 if __name__ == '__main__':
     test_simple_inversion()
@@ -594,3 +805,5 @@ if __name__ == '__main__':
     test_simplify()
     test_evaldomain()
     test_localize()
+    test_prepare_reference_domain()
+    test_extract_models()
