@@ -1,5 +1,5 @@
-#This file is part of Tryton.  The COPYRIGHT file at the top level of
-#this repository contains the full copyright notices and license terms.
+# This file is part of Tryton.  The COPYRIGHT file at the top level of
+# this repository contains the full copyright notices and license terms.
 import operator
 import math
 
@@ -21,17 +21,18 @@ class SelectionMixin(object):
         self._values2selection = {}
         self._domain_cache = {}
 
-    def init_selection(self, key=None):
-        if key is None:
-            key = tuple(sorted((k, None)
-                    for k in self.attrs.get('selection_change_with') or []))
+    def init_selection(self, value=None):
+        if value is None:
+            value = dict((k, None)
+                for k in self.attrs.get('selection_change_with') or [])
+        key = freeze_value(value)
         selection = self.attrs.get('selection', [])[:]
         if (not isinstance(selection, (list, tuple))
                 and key not in self._values2selection):
             try:
                 if self.attrs.get('selection_change_with'):
                     selection = RPCExecute('model', self.model_name, selection,
-                        dict(key))
+                        value)
                 else:
                     selection = RPCExecute('model', self.model_name, selection)
             except RPCException:
@@ -49,21 +50,15 @@ class SelectionMixin(object):
             return
 
         domain = field.domain_get(record)
-        if field.attrs['type'] == 'reference':
-            # The domain on reference field is not only based on the selection
-            # so the selection can not be filtered.
-            domain = []
         if 'relation' not in self.attrs:
             change_with = self.attrs.get('selection_change_with') or []
-            args = record._get_on_change_args(change_with)
-            del args['id']
-            key = args.items()
-            key.sort()
-            self.init_selection(tuple(key))
+            value = record._get_on_change_args(change_with)
+            del value['id']
+            self.init_selection(value)
             self.filter_selection(domain, record, field)
         else:
-            context = record[self.field_name].context_get(record)
-            domain_cache_key = str(domain) + str(context)
+            context = field.get_context(record)
+            domain_cache_key = (freeze_value(domain), freeze_value(context))
             if domain_cache_key in self._domain_cache:
                 self.selection = self._domain_cache[domain_cache_key]
                 self._last_domain = (domain, context)
@@ -91,10 +86,23 @@ class SelectionMixin(object):
     def filter_selection(self, domain, record, field):
         if not domain:
             return
-        test = lambda value: eval_domain(domain, {
-                self.field_name: value[0],
-                })
-        self.selection = filter(test, self.selection)
+
+        def _value_evaluator(value):
+            return eval_domain(domain, {
+                    self.field_name: value[0],
+                    })
+
+        def _model_evaluator(allowed_models):
+            def test(value):
+                return value[0] in allowed_models
+            return test
+
+        if field.attrs['type'] == 'reference':
+            allowed_models = field.get_models(record)
+            evaluator = _model_evaluator(allowed_models)
+        else:
+            evaluator = _value_evaluator
+        self.selection = list(filter(evaluator, self.selection))
 
     def get_inactive_selection(self, value):
         if 'relation' not in self.attrs:
@@ -123,10 +131,21 @@ def selection_shortcuts(entry):
     return entry
 
 
+def freeze_value(value):
+    if isinstance(value, dict):
+        return tuple(sorted((k, freeze_value(v))
+                for k, v in value.items()))
+    elif isinstance(value, (list, set)):
+        return tuple(freeze_value(v) for v in value)
+    else:
+        return value
+
+
 class PopdownMixin(object):
 
     def set_popdown(self, selection, entry):
-        if not entry.child:  # entry is destroyed
+        child = entry.get_child()
+        if not child:  # entry is destroyed
             return
         model, lengths = self.get_popdown_model(selection)
         entry.set_model(model)
@@ -138,7 +157,7 @@ class PopdownMixin(object):
         completion = gtk.EntryCompletion()
         completion.set_inline_selection(True)
         completion.set_model(model)
-        entry.child.set_completion(completion)
+        child.set_completion(completion)
         if lengths:
             pop = sorted(lengths, reverse=True)
             average = sum(pop) / len(pop)
@@ -147,9 +166,9 @@ class PopdownMixin(object):
             width = max(next((x for x in pop if x < (deviation * 4)), 10), 10)
         else:
             width = 10
-        entry.child.set_width_chars(width)
+        child.set_width_chars(width)
         if lengths:
-            entry.child.set_max_length(max(lengths))
+            child.set_max_length(max(lengths))
         completion.set_text_column(0)
         completion.connect('match-selected', self.match_selected, entry)
 
@@ -167,16 +186,19 @@ class PopdownMixin(object):
         model = entry.get_model()
         for i, values in enumerate(model):
             if values[1] == value:
-                entry.set_active(i)
+                gobject.idle_add(entry.set_active, i)
                 break
 
-    def get_popdown_value(self, entry):
+    def get_popdown_value(self, entry, index=1):
         active = entry.get_active()
         if active < 0:
             return None
         else:
             model = entry.get_model()
-            return model[active][1]
+            return model[active][index]
+
+    def get_popdown_text(self, entry):
+        return self.get_popdown_value(entry, index=0)
 
     def set_popdown_value(self, entry, value):
         active = -1
@@ -191,5 +213,12 @@ class PopdownMixin(object):
         entry.set_active(active)
         if active == -1:
             # When setting no item GTK doesn't clear the entry
-            entry.child.set_text('')
+            entry.get_child().set_text('')
         return True
+
+
+def test_freeze_value():
+    assert freeze_value({'foo': 'bar'}) == (('foo', 'bar'),)
+    assert freeze_value([1, 42, 2, 3]) == (1, 42, 2, 3)
+    assert freeze_value('foo') == 'foo'
+    assert freeze_value({'foo': {'bar': 42}}) == (('foo', (('bar', 42),)),)
