@@ -1,9 +1,8 @@
-#This file is part of Tryton.  The COPYRIGHT file at the top level of
-#this repository contains the full copyright notices and license terms.
-from weakref import WeakSet
-
-from record import Record
-from field import Field, O2MField
+# This file is part of Tryton.  The COPYRIGHT file at the top level of
+# this repository contains the full copyright notices and license terms.
+from .record import Record
+from .field import Field, M2OField, ReferenceField
+from tryton import rpc
 from tryton.signal_event import SignalEvent
 from tryton.common.domain_inversion import is_leaf
 from tryton.common import RPCExecute, RPCException, MODELACCESS
@@ -37,7 +36,6 @@ class Group(SignalEvent, list):
         self.__id2record = {}
         self.__field_childs = None
         self.exclude_field = None
-        self.pool = WeakSet()
         self.skip_model_access = False
 
         if self.parent and self.parent.model_name == model_name:
@@ -46,7 +44,7 @@ class Group(SignalEvent, list):
     @property
     def readonly(self):
         # Must skip res.user for Preference windows
-        if (self._context.get('_datetime')
+        if (self.context.get('_datetime')
                 or (not (MODELACCESS[self.model_name]['write']
                         or MODELACCESS[self.model_name]['create'])
                     and not self.skip_model_access)):
@@ -127,10 +125,12 @@ class Group(SignalEvent, list):
         del self.__id2record[record.id]
 
     def clear(self):
-        for record in self[:]:
+        # Use reversed order to minimize the cursor reposition as the cursor
+        # has more chances to be on top of the list.
+        for record in reversed(self[:]):
             self.signal('group-list-changed', ('record-removed', record))
             record.destroy()
-            self.pop(0)
+            self.pop()
         self.__id2record = {}
         self.record_removed, self.record_deleted = [], []
 
@@ -154,26 +154,46 @@ class Group(SignalEvent, list):
         return '<Group %s at %s>' % (self.model_name, id(self))
 
     def load_fields(self, fields):
-        for name, attr in fields.iteritems():
+        for name, attr in fields.items():
             field = Field.get_field(attr['type'])
             attr['name'] = name
             self.fields[name] = field(attr)
-            if isinstance(self.fields[name], O2MField) \
-                    and '_datetime' in self._context:
-                self.fields[name].context.update({
-                    '_datetime': self._context['_datetime'],
-                    })
 
     def save(self):
         saved = []
         for record in self:
             saved.append(record.save(force_reload=False))
         if self.record_deleted:
+            for record in self.record_deleted:
+                self._remove(record)
+                record.destroy()
             self.delete(self.record_deleted)
+            del self.record_deleted[:]
         return saved
 
     def delete(self, records):
-        return Record.delete(records)
+        if not records:
+            return
+        root_group = self.root_group
+        assert all(r.model_name == self.model_name for r in records)
+        assert all(r.group.root_group == root_group for r in records)
+        records = [r for r in records if r.id >= 0]
+        ctx = self.context
+        ctx['_timestamp'] = {}
+        for rec in records:
+            ctx['_timestamp'].update(rec.get_timestamp())
+        record_ids = set(r.id for r in records)
+        reload_ids = set(root_group.on_write_ids(list(record_ids)))
+        reload_ids -= record_ids
+        reload_ids = list(reload_ids)
+        try:
+            RPCExecute('model', self.model_name, 'delete', list(record_ids),
+                context=ctx)
+        except RPCException:
+            return False
+        if reload_ids:
+            root_group.reload(reload_ids)
+        return True
 
     @property
     def root_group(self):
@@ -185,7 +205,7 @@ class Group(SignalEvent, list):
         return root
 
     def written(self, ids):
-        if isinstance(ids, (int, long)):
+        if isinstance(ids, int):
             ids = [ids]
         ids = [x for x in self.on_write_ids(ids) or [] if x not in ids]
         if not ids:
@@ -199,7 +219,7 @@ class Group(SignalEvent, list):
         for id_ in ids:
             record = self.get(id_)
             if record and not record.modified:
-                record._loaded.clear()
+                record.cancel()
 
     def on_write_ids(self, ids):
         if not self.on_write:
@@ -245,28 +265,34 @@ class Group(SignalEvent, list):
             self.signal('group-cleared')
 
         if new_records and modified:
+            for record in new_records:
+                record.modified_fields.setdefault('id')
             new_records[0].signal('record-modified')
             new_records[0].signal('record-changed')
 
         self.current_idx = 0
         return True
 
-    def _get_context(self):
-        ctx = self._context.copy()
+    @property
+    def context(self):
+        ctx = rpc.CONTEXT.copy()
         if self.parent:
-            ctx.update(self.parent.context_get())
+            parent_context = self.parent.get_context()
+            ctx.update(parent_context)
             if self.child_name in self.parent.group.fields:
                 field = self.parent.group.fields[self.child_name]
-                ctx.update(field.context_get(self.parent))
+                ctx.update(field.get_context(self.parent, parent_context))
         ctx.update(self._context)
         if self.parent_datetime_field:
             ctx['_datetime'] = self.parent.get_eval(
                 )[self.parent_datetime_field]
         return ctx
 
-    context = property(_get_context)
+    @context.setter
+    def context(self, value):
+        self._context = value.copy()
 
-    def add(self, record, position=-1):
+    def add(self, record, position=-1, signal=True):
         if record.group is not self:
             record.signal_unconnect(record.group)
             record.group = self
@@ -286,15 +312,44 @@ class Group(SignalEvent, list):
         self.current_idx = position
         record.modified_fields.setdefault('id')
         record.signal('record-modified')
-        self.signal('group-changed', record)
+        if signal:
+            self.signal('group-changed', record)
+            # Set parent field to trigger on_change
+            if self.parent and self.parent_name in self.fields:
+                field = self.fields[self.parent_name]
+                if isinstance(field, (M2OField, ReferenceField)):
+                    value = self.parent.id, ''
+                    if isinstance(field, ReferenceField):
+                        value = self.parent.model_name, value
+                    field.set_client(record, value)
         return record
 
     def set_sequence(self, field='sequence'):
-        index = 0
         changed = False
+        prev = None
         for record in self:
-            if record[field]:
-                if index >= record[field].get(record):
+            # Assume not loaded records are correctly ordered
+            # as far as we do not change any previous records.
+            if record.get_loaded([field]) or changed or record.id < 0:
+                if prev:
+                    index = prev[field].get(prev)
+                else:
+                    index = None
+                update = False
+                value = record[field].get(record)
+                if value is None:
+                    if index:
+                        update = True
+                    elif prev and record.id >= 0:
+                        update = record.id < prev.id
+                elif value == index:
+                    if prev and record.id >= 0:
+                        update = record.id < prev.id
+                elif value <= (index or 0):
+                    update = True
+                if update:
+                    if index is None:
+                        index = 0
                     index += 1
                     record.signal_unconnect(self, 'record-changed')
                     try:
@@ -303,15 +358,14 @@ class Group(SignalEvent, list):
                         record.signal_connect(self, 'record-changed',
                             self._record_changed)
                     changed = record
-                else:
-                    index = record[field].get(record)
+            prev = record
         if changed:
             self.signal('group-changed', changed)
 
-    def new(self, default=True, obj_id=None):
+    def new(self, default=True, obj_id=None, rec_name=None):
         record = Record(self.model_name, obj_id, group=self)
         if default:
-            record.default_get()
+            record.default_get(rec_name=rec_name)
         record.signal_connect(self, 'record-changed', self._record_changed)
         record.signal_connect(self, 'record-modified', self._record_modified)
         return record
@@ -342,7 +396,7 @@ class Group(SignalEvent, list):
         if modified:
             record.modified_fields.setdefault('id')
             record.signal('record-modified')
-        if not record.parent or self[idx].id < 0 or force_remove:
+        if self[idx].id < 0 or force_remove:
             self._remove(self[idx])
 
         if len(self):
@@ -368,7 +422,7 @@ class Group(SignalEvent, list):
             return None
         return self[self.current_idx]
 
-    def next(self):
+    def __next__(self):
         if len(self) and self.current_idx is not None:
             self.current_idx = (self.current_idx + 1) % len(self)
         elif len(self):
@@ -377,9 +431,9 @@ class Group(SignalEvent, list):
             return None
         return self[self.current_idx]
 
-    def add_fields(self, fields, signal=True):
+    def add_fields(self, fields):
         to_add = {}
-        for name, attr in fields.iteritems():
+        for name, attr in fields.items():
             if name not in self.fields:
                 to_add[name] = attr
             else:
@@ -397,11 +451,13 @@ class Group(SignalEvent, list):
         if len(new) and len(to_add):
             try:
                 values = RPCExecute('model', self.model_name, 'default_get',
-                    to_add.keys(), context=self.context)
+                    list(to_add.keys()), context=self.context)
             except RPCException:
                 return False
             for record in new:
-                record.set_default(values, signal=signal)
+                record.set_default(values, signal=False)
+            # Trigger signal only once with the last record
+            record.signal('record-changed')
 
     def get(self, id):
         'Return record with the id'
@@ -419,6 +475,10 @@ class Group(SignalEvent, list):
                 self.parent.group.children.remove(self)
             except ValueError:
                 pass
+        # One2Many connect the group to itself to send signals to the parent
+        # but as we are destroying the group, we do not need to notify the
+        # parent otherwise it will trigger unnecessary display.
+        self.signal_unconnect(self)
         self.clear()
         super(Group, self).destroy()
 
