@@ -1,11 +1,9 @@
-#This file is part of Tryton.  The COPYRIGHT file at the top level of
-#this repository contains the full copyright notices and license terms.
-import tryton.rpc as rpc
+# This file is part of Tryton.  The COPYRIGHT file at the top level of
+# this repository contains the full copyright notices and license terms.
 from tryton.signal_event import SignalEvent
 import tryton.common as common
 from tryton.pyson import PYSONDecoder
-import field as fields
-from functools import reduce
+from . import field as fields
 from tryton.common import RPCExecute, RPCException
 from tryton.config import CONFIG
 
@@ -31,12 +29,13 @@ class Record(SignalEvent):
         self.modified_fields = {}
         self._timestamp = None
         self.attachment_count = -1
+        self.unread_note = -1
+        self.button_clicks = {}
         self.next = {}  # Used in Group list
         self.value = {}
         self.autocompletion = {}
         self.exception = False
         self.destroyed = False
-        self.pool.add(self)
 
     def __getitem__(self, name):
         if name not in self._loaded and self.id >= 0:
@@ -44,26 +43,31 @@ class Record(SignalEvent):
                 self.id: self,
                 }
             if name == '*':
-                loading = reduce(
-                        lambda x, y: 'eager' if x == y == 'eager' else 'lazy',
-                        (field.attrs.get('loading', 'eager')
-                            for field in self.group.fields.itervalues()),
-                        'eager')
+                loading = 'eager'
+                views = set()
+                for field in self.group.fields.values():
+                    if field.attrs.get('loading', 'eager') == 'lazy':
+                        loading = 'lazy'
+                    views |= field.views
                 # Set a valid name for next loaded check
-                for fname, field in self.group.fields.iteritems():
+                for fname, field in self.group.fields.items():
                     if field.attrs.get('loading', 'eager') == loading:
                         name = fname
                         break
             else:
                 loading = self.group.fields[name].attrs.get('loading', 'eager')
+                views = self.group.fields[name].views
 
             if loading == 'eager':
-                fnames = [fname
-                    for fname, field in self.group.fields.iteritems()
-                    if field.attrs.get('loading', 'eager') == 'eager']
+                fields = ((fname, field)
+                    for fname, field in self.group.fields.items()
+                    if field.attrs.get('loading', 'eager') == 'eager')
             else:
-                fnames = self.group.fields.keys()
-            fnames = [fname for fname in fnames if fname not in self._loaded]
+                fields = self.group.fields.items()
+
+            fnames = [fname for fname, field in fields
+                if fname not in self._loaded
+                and (not views or (views & field.views))]
             fnames.extend(('%s.rec_name' % fname for fname in fnames[:]
                     if self.group.fields[fname].attrs['type']
                     in ('many2one', 'one2one', 'reference')))
@@ -71,70 +75,66 @@ class Record(SignalEvent):
                 fnames.append('rec_name')
             fnames.append('_timestamp')
 
-            record_context = self.context_get()
+            record_context = self.get_context()
             if loading == 'eager':
-                limit = CONFIG['client.limit']
-                if not self.parent:
-                    # If not a children no need to load too much
-                    limit = int(limit / len(fnames))
+                limit = int(CONFIG['client.limit'] / len(fnames))
 
                 def filter_group(record):
                     return name not in record._loaded and record.id >= 0
 
-                def filter_pool(record):
+                def filter_parent_group(record):
                     return (filter_group(record)
                         and record.id not in id2record
-                        and record.context_get() == record_context)
+                        and ((record.group == self.group)
+                            # Don't compute context for same group
+                            or (record.get_context() == record_context)))
 
-                if self.parent:
-                    pool = list(self.pool)
+                if self.parent and self.parent.model_name == self.model_name:
+                    group = sum(self.parent.group.children, [])
+                    filter_ = filter_parent_group
                 else:
-                    # Don't look at the pool if it is the root
-                    pool = []
-                for group, filter_ in (
-                        (self.group, filter_group),
-                        (pool, filter_pool),
-                        ):
-                    if self in group:
-                        idx = group.index(self)
-                        length = len(group)
-                        n = 1
-                        while len(id2record) < limit and (idx - n >= 0
-                                or idx + n < length) and n < 2 * limit:
-                            if idx - n >= 0:
-                                record = group[idx - n]
-                                if filter_(record):
-                                    id2record[record.id] = record
-                            if idx + n < length:
-                                record = group[idx + n]
-                                if filter_(record):
-                                    id2record[record.id] = record
-                            n += 1
+                    group = self.group
+                    filter_ = filter_group
+                if self in group:
+                    idx = group.index(self)
+                    length = len(group)
+                    n = 1
+                    while len(id2record) < limit and (idx - n >= 0
+                            or idx + n < length) and n < 2 * limit:
+                        if idx - n >= 0:
+                            record = group[idx - n]
+                            if filter_(record):
+                                id2record[record.id] = record
+                        if idx + n < length:
+                            record = group[idx + n]
+                            if filter_(record):
+                                id2record[record.id] = record
+                        n += 1
 
             ctx = record_context.copy()
             ctx.update(dict(('%s.%s' % (self.model_name, fname), 'size')
-                    for fname, field in self.group.fields.iteritems()
+                    for fname, field in self.group.fields.items()
                     if field.attrs['type'] == 'binary' and fname in fnames))
-            exception = None
+            exception = False
             try:
                 values = RPCExecute('model', self.model_name, 'read',
-                    id2record.keys(), fnames, context=ctx)
-            except RPCException, exception:
+                    list(id2record.keys()), fnames, context=ctx)
+            except RPCException:
                 values = [{'id': x} for x in id2record]
                 default_values = dict((f, None) for f in fnames)
                 for value in values:
                     value.update(default_values)
-                self.exception = True
+                self.exception = exception = True
             id2value = dict((value['id'], value) for value in values)
-            for id, record in id2record.iteritems():
+            for id, record in id2record.items():
                 if not record.exception:
-                    record.exception = bool(exception)
+                    record.exception = exception
                 value = id2value.get(id)
                 if record and not record.destroyed and value:
                     for key in record.modified_fields:
                         value.pop(key, None)
                     record.set(value, signal=False)
-        return self.group.fields.get(name, False)
+        return self.group.fields[name]
 
     def __repr__(self):
         return '<Record %s@%s at %s>' % (self.id, self.model_name, id(self))
@@ -143,14 +143,6 @@ class Record(SignalEvent):
     def modified(self):
         return bool(self.modified_fields)
 
-    @property
-    def pool(self):
-        group = self.group
-        while (group.parent is not None
-                and group.parent.model_name == self.model_name):
-            group = group.parent
-        return group.pool
-
     @property
     def parent(self):
         return self.group.parent
@@ -208,6 +200,17 @@ class Record(SignalEvent):
         path.reverse()
         return tuple(path)
 
+    def get_index_path(self, group=None):
+        path = []
+        record = self
+        while record:
+            path.append(record.group.index(record))
+            if record.group is group:
+                break
+            record = record.parent
+        path.reverse()
+        return tuple(path)
+
     def get_removed(self):
         if self.group is not None:
             return self in self.group.record_removed
@@ -250,27 +253,35 @@ class Record(SignalEvent):
 
     def get(self):
         value = {}
-        for name, field in self.group.fields.iteritems():
-            if field.attrs.get('readonly'):
+        for name, field in self.group.fields.items():
+            if (field.attrs.get('readonly')
+                    and not isinstance(field, fields.O2MField)):
                 continue
             if field.name not in self.modified_fields and self.id >= 0:
                 continue
             value[name] = field.get(self)
+            # Sending an empty x2MField breaks ModelFieldAccess.check
+            if isinstance(field, fields.O2MField) and not value[name]:
+                del value[name]
         return value
 
     def get_eval(self):
         value = {}
-        for name, field in self.group.fields.iteritems():
+        for name, field in self.group.fields.items():
             if name not in self._loaded and self.id >= 0:
                 continue
             value[name] = field.get_eval(self)
         value['id'] = self.id
         return value
 
-    def get_on_change_value(self):
+    def get_on_change_value(self, skip=None):
         value = {}
-        for name, field in self.group.fields.iteritems():
-            if name not in self._loaded and self.id >= 0:
+        for name, field in self.group.fields.items():
+            if skip and name in skip:
+                continue
+            if (self.id >= 0
+                    and (name not in self._loaded
+                        or name not in self.modified_fields)):
                 continue
             value[name] = field.get_on_change_value(self)
         value['id'] = self.id
@@ -279,10 +290,12 @@ class Record(SignalEvent):
     def cancel(self):
         self._loaded.clear()
         self.modified_fields.clear()
+        self._timestamp = None
+        self.button_clicks.clear()
 
     def get_timestamp(self):
         result = {self.model_name + ',' + str(self.id): self._timestamp}
-        for name, field in self.group.fields.iteritems():
+        for name, field in self.group.fields.items():
             if name in self._loaded:
                 result.update(field.get_timestamp(self))
         return result
@@ -293,7 +306,7 @@ class Record(SignalEvent):
         values = self._get_on_change_args(self.modified_fields)
         try:
             RPCExecute('model', self.model_name, 'pre_validate', values,
-                context=self.context_get())
+                context=self.get_context())
         except RPCException:
             return False
         return True
@@ -304,7 +317,7 @@ class Record(SignalEvent):
                 value = self.get()
                 try:
                     res, = RPCExecute('model', self.model_name, 'create',
-                        [value], context=self.context_get())
+                        [value], context=self.get_context())
                 except RPCException:
                     return False
                 old_id = self.id
@@ -313,7 +326,7 @@ class Record(SignalEvent):
             elif self.modified:
                 value = self.get()
                 if value:
-                    context = self.context_get()
+                    context = self.get_context()
                     context = context.copy()
                     context['_timestamp'] = self.get_timestamp()
                     try:
@@ -321,8 +334,7 @@ class Record(SignalEvent):
                             [self.id], value, context=context)
                     except RPCException:
                         return False
-            self._loaded.clear()
-            self.modified_fields = {}
+            self.cancel()
             if force_reload:
                 self.reload()
             if self.group:
@@ -332,37 +344,13 @@ class Record(SignalEvent):
             self.parent.save(force_reload=force_reload)
         return self.id
 
-    @staticmethod
-    def delete(records):
-        if not records:
-            return
-        record = records[0]
-        root_group = record.group.root_group
-        assert all(r.model_name == record.model_name for r in records)
-        assert all(r.group.root_group == root_group for r in records)
-        records = [r for r in records if r.id >= 0]
-        ctx = {}
-        ctx['_timestamp'] = {}
-        for rec in records:
-            ctx['_timestamp'].update(rec.get_timestamp())
-        record_ids = set(r.id for r in records)
-        reload_ids = set(root_group.on_write_ids(list(record_ids)))
-        reload_ids -= record_ids
-        reload_ids = list(reload_ids)
-        try:
-            RPCExecute('model', record.model_name, 'delete', list(record_ids),
-                context=ctx)
-        except RPCException:
-            return False
-        if reload_ids:
-            root_group.reload(reload_ids)
-        return True
-
-    def default_get(self):
+    def default_get(self, rec_name=None):
         if len(self.group.fields):
+            context = self.get_context()
+            context.setdefault('default_rec_name', rec_name)
             try:
                 vals = RPCExecute('model', self.model_name, 'default_get',
-                    self.group.fields.keys(), context=self.context_get())
+                    list(self.group.fields.keys()), context=context)
             except RPCException:
                 return
             if (self.parent
@@ -375,50 +363,48 @@ class Record(SignalEvent):
                         == self.group.parent.model_name):
                     vals[self.parent_name] = self.parent.id
             self.set_default(vals)
-        for fieldname, fieldinfo in self.group.fields.iteritems():
-            if not fieldinfo.attrs.get('autocomplete'):
-                continue
-            self.do_autocomplete(fieldname)
         return vals
 
     def rec_name(self):
         try:
             return RPCExecute('model', self.model_name, 'read', [self.id],
-                ['rec_name'], context=self.context_get())[0]['rec_name']
+                ['rec_name'], context=self.get_context())[0]['rec_name']
         except RPCException:
             return ''
 
-    def validate(self, fields=None, softvalidation=False):
+    def validate(self, fields=None, softvalidation=False, pre_validate=None):
         if isinstance(fields, list) and fields:
             self._check_load(fields)
         elif fields is None:
             self._check_load()
         res = True
-        for field_name, field in self.group.fields.iteritems():
+        for field_name, field in self.group.fields.items():
             if fields is not None and field_name not in fields:
                 continue
-            if field.get_state_attrs(self).get('readonly', False):
+            if field.attrs.get('readonly'):
                 continue
             if field_name == self.group.exclude_field:
                 continue
-            if not field.validate(self, softvalidation):
+            if not field.validate(self, softvalidation, pre_validate):
                 res = False
         return res
 
     def _get_invalid_fields(self):
-        res = []
-        for fname, field in self.group.fields.iteritems():
-            if not field.get_state_attrs(self).get('valid', True):
-                res.append((fname, field.attrs['string']))
-        return dict(res)
+        fields = {}
+        for fname, field in self.group.fields.items():
+            invalid = field.get_state_attrs(self).get('invalid')
+            if invalid:
+                fields[fname] = invalid
+        return fields
 
     invalid_fields = property(_get_invalid_fields)
 
-    def context_get(self):
+    def get_context(self):
         return self.group.context
 
-    def set_default(self, val, signal=True):
-        for fieldname, value in val.items():
+    def set_default(self, val, signal=True, validate=True):
+        fieldnames = []
+        for fieldname, value in list(val.items()):
             if fieldname not in self.group.fields:
                 continue
             if fieldname == self.group.exclude_field:
@@ -432,15 +418,22 @@ class Record(SignalEvent):
                     del self.value[field_rec_name]
             self.group.fields[fieldname].set_default(self, value)
             self._loaded.add(fieldname)
-        self.validate(softvalidation=True)
+            fieldnames.append(fieldname)
+        self.on_change(fieldnames)
+        self.on_change_with(fieldnames)
+        if validate:
+            self.validate(softvalidation=True)
         if signal:
             self.signal('record-changed')
 
-    def set(self, val, signal=True):
+    def set(self, val, signal=True, validate=True):
         later = {}
-        for fieldname, value in val.iteritems():
+        fieldnames = []
+        for fieldname, value in val.items():
             if fieldname == '_timestamp':
-                self._timestamp = value
+                # Always keep the older timestamp
+                if not self._timestamp:
+                    self._timestamp = value
                 continue
             if fieldname not in self.group.fields:
                 if fieldname == 'rec_name':
@@ -458,20 +451,19 @@ class Record(SignalEvent):
                     del self.value[field_rec_name]
             self.group.fields[fieldname].set(self, value)
             self._loaded.add(fieldname)
-        for fieldname, value in later.iteritems():
+            fieldnames.append(fieldname)
+        for fieldname, value in later.items():
             self.group.fields[fieldname].set(self, value)
             self._loaded.add(fieldname)
+        if validate:
+            self.validate(fieldnames, softvalidation=True)
         if signal:
             self.signal('record-changed')
 
     def set_on_change(self, values):
-        later = {}
-        for fieldname, value in values.items():
+        for fieldname, value in list(values.items()):
             if fieldname not in self.group.fields:
                 continue
-            if isinstance(self.group.fields[fieldname], fields.O2MField):
-                later[fieldname] = value
-                continue
             if isinstance(self.group.fields[fieldname], (fields.M2OField,
                         fields.ReferenceField)):
                 field_rec_name = fieldname + '.rec_name'
@@ -480,14 +472,6 @@ class Record(SignalEvent):
                 elif field_rec_name in self.value:
                     del self.value[field_rec_name]
             self.group.fields[fieldname].set_on_change(self, value)
-        for fieldname, value in later.items():
-            # on change recursion checking is done only for x2many
-            field_x2many = self.group.fields[fieldname]
-            try:
-                field_x2many.in_on_change = True
-                field_x2many.set_on_change(self, value)
-            finally:
-                field_x2many.in_on_change = False
 
     def reload(self, fields=None):
         if self.id < 0:
@@ -499,16 +483,29 @@ class Record(SignalEvent):
                 self[field]
         self.validate(fields or [])
 
+    def reset(self, value):
+        self.cancel()
+        self.set(value, signal=False)
+
+        if self.parent:
+            self.parent.on_change([self.group.child_name])
+            self.parent.on_change_with([self.group.child_name])
+
+        self.signal('record-changed')
+
     def expr_eval(self, expr):
-        if not isinstance(expr, basestring):
+        if not isinstance(expr, str):
             return expr
-        ctx = rpc.CONTEXT.copy()
-        ctx['context'] = ctx.copy()
-        ctx['context'].update(self.context_get())
-        ctx.update(self.get_eval())
+        if not expr:
+            return
+        elif expr == '[]':
+            return []
+        elif expr == '{}':
+            return {}
+        ctx = self.get_eval()
+        ctx['context'] = self.get_context()
         ctx['active_model'] = self.model_name
         ctx['active_id'] = self.id
-        ctx['_user'] = rpc._USER
         if self.parent and self.parent_name:
             ctx['_parent_' + self.parent_name] = \
                 common.EvalEnvironment(self.parent)
@@ -529,18 +526,25 @@ class Record(SignalEvent):
         res['id'] = self.id
         return res
 
-    def on_change(self, fieldname, attr):
-        if isinstance(attr, basestring):
-            attr = PYSONDecoder().decode(attr)
-        args = self._get_on_change_args(attr)
-        try:
-            res = RPCExecute('model', self.model_name, 'on_change_' +
-                fieldname, args, context=self.context_get())
-        except RPCException:
-            return
-        self.set_on_change(res)
+    def on_change(self, fieldnames):
+        values = {}
+        for fieldname in fieldnames:
+            on_change = self.group.fields[fieldname].attrs.get('on_change')
+            if not on_change:
+                continue
+            values.update(self._get_on_change_args(on_change))
 
-    def on_change_with(self, field_name):
+        if values:
+            try:
+                changes = RPCExecute('model', self.model_name, 'on_change',
+                    values, fieldnames, context=self.get_context())
+            except RPCException:
+                return
+            for change in changes:
+                self.set_on_change(change)
+
+    def on_change_with(self, field_names):
+        field_names = set(field_names)
         fieldnames = set()
         values = {}
         later = set()
@@ -549,9 +553,7 @@ class Record(SignalEvent):
                     'on_change_with')
             if not on_change_with:
                 continue
-            if field_name not in on_change_with:
-                continue
-            if field_name == fieldname:
+            if not field_names & set(on_change_with):
                 continue
             if fieldnames & set(on_change_with):
                 later.add(fieldname)
@@ -566,7 +568,7 @@ class Record(SignalEvent):
         if fieldnames:
             try:
                 result = RPCExecute('model', self.model_name, 'on_change_with',
-                    values, list(fieldnames), context=self.context_get())
+                    values, list(fieldnames), context=self.get_context())
             except RPCException:
                 return
             self.set_on_change(result)
@@ -577,13 +579,13 @@ class Record(SignalEvent):
             try:
                 result = RPCExecute('model', self.model_name,
                     'on_change_with_' + fieldname, values,
-                    context=self.context_get())
+                    context=self.get_context())
             except RPCException:
                 return
             self.group.fields[fieldname].set_on_change(self, result)
 
     def autocomplete_with(self, field_name):
-        for fieldname, fieldinfo in self.group.fields.iteritems():
+        for fieldname, fieldinfo in self.group.fields.items():
             autocomplete = fieldinfo.attrs.get('autocomplete', [])
             if field_name not in autocomplete:
                 continue
@@ -595,12 +597,22 @@ class Record(SignalEvent):
         args = self._get_on_change_args(autocomplete)
         try:
             res = RPCExecute('model', self.model_name, 'autocomplete_' +
-                fieldname, args, context=self.context_get())
+                fieldname, args, context=self.get_context())
         except RPCException:
             # ensure res is a list
             res = []
         self.autocompletion[fieldname] = res
 
+    def set_field_context(self):
+        from .group import Group
+        for name, field in self.group.fields.items():
+            value = self.value.get(name)
+            if not isinstance(value, Group):
+                continue
+            context = field.attrs.get('context')
+            if context:
+                value.context = self.expr_eval(context)
+
     def get_attachment_count(self, reload=False):
         if self.id < 0:
             return 0
@@ -610,17 +622,43 @@ class Record(SignalEvent):
                     'search_count', [
                         ('resource', '=',
                             '%s,%s' % (self.model_name, self.id)),
-                        ], context=self.context_get())
+                        ], context=self.get_context())
             except RPCException:
                 return 0
         return self.attachment_count
 
+    def get_unread_note(self, reload=False):
+        if self.id < 0:
+            return 0
+        if self.unread_note < 0 or reload:
+            try:
+                self.unread_note = RPCExecute('model', 'ir.note',
+                    'search_count', [
+                        ('resource', '=',
+                            '%s,%s' % (self.model_name, self.id)),
+                        ('unread', '=', True),
+                        ], context=self.get_context())
+            except RPCException:
+                return 0
+        return self.unread_note
+
+    def get_button_clicks(self, name):
+        if self.id < 0:
+            return
+        clicks = self.button_clicks.get(name)
+        if clicks is not None:
+            return clicks
+        try:
+            clicks = RPCExecute('model', 'ir.model.button.click',
+                'get_click', self.model_name, name, self.id)
+            self.button_clicks[name] = clicks
+        except RPCException:
+            return
+        return clicks
+
     def destroy(self):
-        # Get reference to pool before unref group
-        pool = self.pool
-        for v in self.value.itervalues():
+        for v in self.value.values():
             if hasattr(v, 'destroy'):
                 v.destroy()
         super(Record, self).destroy()
         self.destroyed = True
-        pool.remove(self)
