@@ -1,8 +1,12 @@
-#This file is part of Tryton.  The COPYRIGHT file at the top level of
-#this repository contains the full copyright notices and license terms.
+# This file is part of Tryton.  The COPYRIGHT file at the top level of
+# this repository contains the full copyright notices and license terms.
+import logging
+import gettext
+
 import gtk
 import pango
-import gettext
+
+from gi.repository import Gtk
 
 from tryton.signal_event import SignalEvent
 import tryton.common as common
@@ -11,14 +15,18 @@ from tryton.gui import Main
 from tryton.exceptions import TrytonServerError
 from tryton.gui.window.nomodal import NoModal
 from tryton.common.button import Button
-from tryton.common import RPCExecute, RPCException
+from tryton.common import RPCExecute, RPCException, RPCContextReload
 from tryton.common import TRYTON_ICON
+from .infobar import InfoBar
+from .tabcontent import TabContent
+
 _ = gettext.gettext
+logger = logging.getLogger(__name__)
 
 
-class Wizard(object):
+class Wizard(InfoBar):
 
-    def __init__(self, name=False):
+    def __init__(self, name=''):
         super(Wizard, self).__init__()
         self.widget = gtk.VBox(spacing=3)
         self.toolbar_box = None
@@ -53,7 +61,11 @@ class Wizard(object):
         self.direct_print = direct_print
         self.email_print = email_print
         self.email = email
-        self.context = context
+        self.context = context.copy() if context is not None else {}
+        self.context['active_id'] = self.id
+        self.context['active_ids'] = self.ids
+        self.context['active_model'] = self.model
+        self.context['action_id'] = self.action_id
 
         def callback(result):
             try:
@@ -73,10 +85,6 @@ class Wizard(object):
         self.__processing = True
 
         ctx = self.context.copy()
-        ctx['active_id'] = self.id
-        ctx['active_ids'] = self.ids
-        ctx['active_model'] = self.model
-        ctx['action_id'] = self.action_id
         if self.screen:
             data = {
                 self.screen_state: self.screen.get_on_change_value(),
@@ -87,7 +95,7 @@ class Wizard(object):
         def callback(result):
             try:
                 result = result()
-            except RPCException, rpc_exception:
+            except RPCException as rpc_exception:
                 if (not isinstance(rpc_exception.exception,
                         TrytonServerError)
                         or not self.screen):
@@ -99,8 +107,13 @@ class Wizard(object):
             if 'view' in result:
                 self.clean()
                 view = result['view']
-                self.update(view['fields_view'], view['defaults'],
-                    view['buttons'])
+                self.update(view['fields_view'], view['buttons'])
+
+                self.screen.new(default=False)
+                self.screen.current_record.set_default(view['defaults'])
+                self.update_buttons(self.screen.current_record)
+                self.screen.set_cursor()
+
                 self.screen_state = view['state']
                 self.__waiting_response = True
             else:
@@ -108,7 +121,19 @@ class Wizard(object):
 
             def execute_actions():
                 for action in result.get('actions', []):
-                    Action._exec_action(*action, context=self.context.copy())
+                    for k, v in [
+                            ('direct_print', self.direct_print),
+                            ('email_print', self.email_print),
+                            ('email', self.email),
+                            ]:
+                        action[0].setdefault(k, v)
+                    context = self.context.copy()
+                    # Remove wizard keys added by run
+                    del context['active_id']
+                    del context['active_ids']
+                    del context['active_model']
+                    del context['action_id']
+                    Action._exec_action(*action, context=context)
 
             if self.state == self.end_state:
                 self.end(lambda *a: execute_actions())
@@ -119,13 +144,22 @@ class Wizard(object):
         RPCExecute('wizard', self.action, 'execute', self.session_id, data,
             self.state, context=ctx, callback=callback)
 
-    def destroy(self):
+    def destroy(self, action=None):
         if self.screen:
             self.screen.destroy()
 
     def end(self, callback=None):
-        RPCExecute('wizard', self.action, 'delete', self.session_id,
-            process_exception=False, callback=callback)
+        def end_callback(action):
+            self.destroy(action=action())
+            if callback:
+                callback()
+        try:
+            RPCExecute('wizard', self.action, 'delete', self.session_id,
+                process_exception=False, callback=end_callback)
+        except Exception:
+            logger.warn(
+                _("Unable to delete wizard %s") % self.session_id,
+                exc_info=True)
 
     def clean(self):
         for widget in self.widget.get_children():
@@ -139,7 +173,9 @@ class Wizard(object):
         if (not self.screen.current_record.validate()
                 and state != self.end_state):
             self.screen.display(set_cursor=True)
+            self.message_info(self.screen.invalid_message(), gtk.MESSAGE_ERROR)
             return
+        self.message_info()
         self.state = state
         self.process()
 
@@ -151,12 +187,15 @@ class Wizard(object):
         button.show()
         return button
 
-    def _record_modified(self, screen, signal):
-        record = screen.current_record
-        for button in self.states.itervalues():
+    def _record_changed(self, screen, record):
+        self.update_buttons(record)
+
+    def update_buttons(self, record):
+        for button in self.states.values():
             button.state_set(record)
 
-    def update(self, view, defaults, buttons):
+    def update(self, view, buttons):
+        tooltips = common.Tooltips()
         for button in buttons:
             self._get_button(button)
 
@@ -164,34 +203,23 @@ class Wizard(object):
         self.screen.add_view(view)
         self.screen.switch_view()
         self.screen.widget.show()
-        self.screen.signal_connect(self, 'record-modified',
-            self._record_modified)
+        self.screen.signal_connect(self, 'group-changed',
+            self._record_changed)
 
         title = gtk.Label()
         title.modify_font(pango.FontDescription("bold 14"))
-        title.set_label(self.screen.current_view.title)
+        title.set_label(common.ellipsize(self.name, 80))
+        tooltips.set_tip(title, self.name)
         title.set_padding(20, 4)
         title.set_alignment(0.0, 0.5)
+        title.set_max_width_chars(1)
+        title.set_ellipsize(pango.ELLIPSIZE_END)
         title.set_size_request(0, -1)  # Allow overflow
         title.modify_fg(gtk.STATE_NORMAL, gtk.gdk.color_parse("#000000"))
         title.show()
 
-        self.info_label = gtk.Label()
-        self.info_label.set_padding(3, 3)
-        self.info_label.set_alignment(1.0, 0.5)
-
-        self.eb_info = gtk.EventBox()
-        self.eb_info.add(self.info_label)
-        self.eb_info.connect('button-release-event',
-                lambda *a: self.message_info(''))
-
-        vbox = gtk.VBox()
-        vbox.pack_start(self.eb_info, expand=True, fill=True, padding=5)
-        vbox.show()
-
         hbox = gtk.HBox()
         hbox.pack_start(title, expand=True, fill=True)
-        hbox.pack_start(vbox, expand=False, fill=True, padding=20)
         hbox.show()
 
         frame = gtk.Frame()
@@ -220,17 +248,16 @@ class Wizard(object):
         self.scrolledwindow.add(viewport)
         self.scrolledwindow.show()
 
-        self.widget.pack_start(self.scrolledwindow)
+        self.widget.pack_start(self.scrolledwindow, expand=True, fill=True)
 
-        self.screen.new(default=False)
-        self.screen.current_record.set_default(defaults)
-        self.screen.set_cursor()
+        self.create_info_bar()
+        self.widget.pack_start(self.info_bar, False, True)
 
 
-class WizardForm(Wizard, SignalEvent):
+class WizardForm(Wizard, TabContent, SignalEvent):
     "Wizard"
 
-    def __init__(self, name=False):
+    def __init__(self, name=''):
         super(WizardForm, self).__init__(name=name)
         self.toolbar_box = gtk.HBox()
         self.hbuttonbox = gtk.HButtonBox()
@@ -256,8 +283,8 @@ class WizardForm(Wizard, SignalEvent):
         self.hbuttonbox.pack_start(button)
         return button
 
-    def update(self, view, defaults, buttons):
-        super(WizardForm, self).update(view, defaults, buttons)
+    def update(self, view, buttons):
+        super(WizardForm, self).update(view, buttons)
         self.widget.pack_start(self.hbuttonbox, expand=False, fill=True)
 
     def sig_close(self):
@@ -265,15 +292,19 @@ class WizardForm(Wizard, SignalEvent):
             self.states[self.end_state].clicked()
         return self.state == self.end_state
 
-    def destroy(self):
+    def destroy(self, action=None):
         if self.toolbar_box.get_children():
             toolbar = self.toolbar_box.get_children()[0]
             self.toolbar_box.remove(toolbar)
-        super(WizardForm, self).destroy()
+        super(WizardForm, self).destroy(action=action)
+        if action == 'reload menu':
+            RPCContextReload(Main().sig_win_menu)
+        elif action == 'reload context':
+            RPCContextReload()
 
     def end(self, callback=None):
         super(WizardForm, self).end(callback=callback)
-        Main.get_main()._win_del(self.widget)
+        Main()._win_del(self.widget)
 
     def set_cursor(self):
         if self.screen:
@@ -282,23 +313,25 @@ class WizardForm(Wizard, SignalEvent):
 
 class WizardDialog(Wizard, NoModal):
 
-    def __init__(self, name=False):
+    def __init__(self, name=''):
         if not name:
             name = _('Wizard')
         Wizard.__init__(self, name=name)
         NoModal.__init__(self)
         self.dia = gtk.Dialog(self.name, self.parent,
             gtk.DIALOG_DESTROY_WITH_PARENT)
+        Main().add_window(self.dia)
         self.dia.set_position(gtk.WIN_POS_CENTER_ON_PARENT)
         self.dia.set_icon(TRYTON_ICON)
         self.dia.set_deletable(False)
+        self.dia.connect('delete-event', lambda *a: True)
         self.dia.connect('close', self.close)
         self.dia.connect('response', self.response)
 
         self.accel_group = gtk.AccelGroup()
         self.dia.add_accel_group(self.accel_group)
 
-        self.dia.vbox.add(self.widget)
+        self.dia.vbox.pack_start(self.widget, expand=True, fill=True)
 
         self.register()
 
@@ -313,28 +346,30 @@ class WizardDialog(Wizard, NoModal):
         response = len(self.states)
         self.dia.add_action_widget(button, response)
         if definition['default']:
-            button.set_flags(gtk.CAN_DEFAULT)
             button.add_accelerator('clicked', self.accel_group,
                 gtk.keysyms.Return, gtk.gdk.CONTROL_MASK,
                 gtk.ACCEL_VISIBLE)
+            button.get_style_context().add_class(
+                Gtk.STYLE_CLASS_SUGGESTED_ACTION)
+            button.set_can_default(True)
+            button.grab_default()
             self.dia.set_default_response(response)
         return button
 
-    def update(self, view, defaults, buttons):
-        # Dialog must be shown before the screen is displayed
-        # to get the treeview realized when displayed
-        sensible_allocation = self.sensible_widget.get_allocation()
-        self.dia.set_default_size(int(sensible_allocation.width * 0.9),
-            int(sensible_allocation.height * 0.9))
-        self.dia.show()
-        common.center_window(self.dia, self.parent, self.sensible_widget)
-        super(WizardDialog, self).update(view, defaults, buttons)
+    def update(self, view, buttons):
+        super(WizardDialog, self).update(view, buttons)
+        current_view = self.screen.current_view
+        self.scrolledwindow.set_policy(gtk.POLICY_NEVER, gtk.POLICY_NEVER)
+        if current_view.scroll:
+            current_view.scroll.set_policy(
+                gtk.POLICY_NEVER, gtk.POLICY_NEVER)
+        self.show()
 
     def destroy(self, action=None):
-        super(WizardDialog, self).destroy()
+        super(WizardDialog, self).destroy(action=action)
         self.dia.destroy()
         NoModal.destroy(self)
-        main = Main.get_main()
+        main = Main()
         if self.parent == main.window:
             current_form = main.get_page()
             if current_form:
@@ -344,24 +379,20 @@ class WizardDialog(Wizard, NoModal):
             dialog = self.page.dialogs[-1]
         else:
             dialog = self.page
-        if (getattr(dialog, 'screen', None)
-                and dialog.screen.current_record
-                and self.sensible_widget != main.window):
-            if dialog.screen.model_name == self.model:
-                ids = self.ids
-            else:
-                # Wizard run from a children record so reload parent record
-                ids = [dialog.screen.current_record.id]
-            dialog.screen.reload(ids, written=True)
+        screen = getattr(dialog, 'screen', None)
+        if self.sensible_widget == main.window:
+            screen = main.menu_screen
+        if screen:
+            if (screen.current_record
+                    and self.sensible_widget != main.window):
+                if screen.model_name == self.model:
+                    ids = self.ids
+                else:
+                    # Wizard run from a children record so reload parent record
+                    ids = [screen.current_record.id]
+                screen.reload(ids, written=True)
             if action:
-                dialog.screen.client_action(action)
-
-    def end(self, callback=None):
-        def end_callback(action):
-            self.destroy(action=action())
-            if callback:
-                callback()
-        super(WizardDialog, self).end(callback=end_callback)
+                screen.client_action(action)
 
     def close(self, widget, event=None):
         widget.emit_stop_by_name('close')
@@ -370,6 +401,7 @@ class WizardDialog(Wizard, NoModal):
         return True
 
     def show(self):
+        self.dia.set_default_size(200, -1)
         self.dia.show()
 
     def hide(self):
